@@ -28,8 +28,21 @@ from typing import (
     get_type_hints,
 )
 
-from .errors import DeserializeError
-from .typedefs import NamedTupleType, NoResult, Possible, T, is_no_result
+from .errors import DepthContainer, DeserializeError
+from .typedefs import (
+    NO_RESULT,
+    UNDEFINED,
+    NamedTupleType,
+    PossibleResult,
+    Undefined,
+)
+
+T = TypeVar("T")  # pylint: disable=invalid-name
+
+
+def _is_no_result(obj: Any) -> bool:
+    """Helper function to determine whether a value is NO_RESULT."""
+    return obj is NO_RESULT
 
 
 def load(
@@ -84,16 +97,19 @@ class Deserialize(Generic[T]):  # pylint: disable=too-many-instance-attributes
 
     obj: Any
     constructor: Type[T]
-    depth: InitVar[List[Type]]
+    depth: InitVar[List[DepthContainer]]
     convert_primitives: bool
-    new_depth: List[Type] = field(init=False)
+    key: Any = UNDEFINED
+    new_depth: List[DepthContainer] = field(init=False)
     constructor_args: Tuple[Type, ...] = field(init=False)
     constructor_origin: Type = field(init=False)
-    check_functions: Iterable[Callable[[], Possible[T]]] = field(init=False)
+    check_functions: Iterable[Callable[[], PossibleResult[T]]] = field(
+        init=False
+    )
 
     def __post_init__(self, depth) -> None:
         """Initialize the uninitialized."""
-        self.new_depth = depth + [self.constructor]
+        self.new_depth = depth + [DepthContainer(self.constructor, self.obj)]
         self.constructor_args = get_args(self.constructor)
         origin = get_origin(self.constructor)
         self.constructor_origin = origin if origin else self.constructor
@@ -101,6 +117,7 @@ class Deserialize(Generic[T]):  # pylint: disable=too-many-instance-attributes
             self._check_any,
             self._check_primitive,
             self._check_none,
+            self._check_undefined,
             self._check_dataclass,
             self._check_namedtuple,
             self._check_typed_dict,
@@ -117,31 +134,46 @@ class Deserialize(Generic[T]):  # pylint: disable=too-many-instance-attributes
         try:
             return next(
                 itertools.dropwhile(
-                    is_no_result,
+                    _is_no_result,
                     (function() for function in self.check_functions),
                 )
             )  # type: ignore
         except StopIteration:
+            # pylint: disable=raise-missing-from
             raise DeserializeError(
                 self.constructor,
                 self.obj,
                 self.new_depth,
+                self.key,
                 message_prefix="Unsupported type. ",
             )
+        except Exception as error:
+            if not isinstance(error, DeserializeError):
+                raise DeserializeError(
+                    self.constructor,
+                    self.obj,
+                    self.new_depth,
+                    self.key,
+                    message_override=str(error),
+                ) from error
+            raise
 
-    def _check_dataclass(self) -> Possible[T]:
+    def _check_dataclass(self) -> PossibleResult[T]:
         """Checks whether a result is a dataclass."""
         if is_dataclass(self.constructor):
             if not isinstance(self.obj, Mapping):
-                raise DeserializeError(Mapping, self.obj, self.new_depth)
+                raise DeserializeError(
+                    Mapping, self.obj, self.new_depth, self.key
+                )
             parameters = inspect.signature(self.constructor).parameters
             return self.constructor(
                 **{
                     name: Deserialize(
-                        obj=self.obj.get(name),
+                        obj=self.obj.get(name, UNDEFINED),
                         constructor=_type,
                         depth=self.new_depth,
                         convert_primitives=self.convert_primitives,
+                        key=name,
                     ).run()
                     for name, _type in get_type_hints(self.constructor).items()
                     if not (
@@ -151,21 +183,24 @@ class Deserialize(Generic[T]):  # pylint: disable=too-many-instance-attributes
                     )
                 }
             )  # type: ignore
-        return NoResult
+        return NO_RESULT
 
-    def _check_namedtuple(self) -> Possible[T]:
+    def _check_namedtuple(self) -> PossibleResult[T]:
         """Checks whether a result is a namedtuple."""
         if isinstance(self.constructor, NamedTupleType):
             if not isinstance(self.obj, Mapping):
-                raise DeserializeError(Mapping, self.obj, self.new_depth)
+                raise DeserializeError(
+                    Mapping, self.obj, self.new_depth, self.key
+                )
             parameters = inspect.signature(self.constructor).parameters
             return self.constructor(
                 **{
                     name: Deserialize(
-                        obj=self.obj.get(name),
+                        obj=self.obj.get(name, UNDEFINED),
                         constructor=_type,
                         depth=self.new_depth,
                         convert_primitives=self.convert_primitives,
+                        key=name,
                     ).run()
                     for name, _type in get_type_hints(self.constructor).items()
                     if not (
@@ -175,15 +210,17 @@ class Deserialize(Generic[T]):  # pylint: disable=too-many-instance-attributes
                     )
                 }
             )  # type: ignore
-        return NoResult
+        return NO_RESULT
 
-    def _check_tuple(self) -> Possible[T]:
+    def _check_tuple(self) -> PossibleResult[T]:
         """Checks whether a result is a Tuple type."""
         if isinstance(self.constructor_origin, type) and issubclass(
             self.constructor_origin, tuple
         ):
             if not isinstance(self.obj, Sequence):
-                raise DeserializeError(tuple, self.obj, self.new_depth)
+                raise DeserializeError(
+                    tuple, self.obj, self.new_depth, self.key
+                )
             if not self.constructor_args:
                 return self.constructor_origin(self.obj)  # type: ignore
             if (
@@ -204,6 +241,7 @@ class Deserialize(Generic[T]):  # pylint: disable=too-many-instance-attributes
                     tuple,
                     self.obj,
                     self.new_depth,
+                    self.key,
                     message_prefix="Tuple incorrect length. ",
                 )
             return self.constructor_origin(
@@ -215,9 +253,9 @@ class Deserialize(Generic[T]):  # pylint: disable=too-many-instance-attributes
                 ).run()
                 for i, arg in enumerate(self.constructor_args)
             )  # type: ignore
-        return NoResult
+        return NO_RESULT
 
-    def _check_sequence(self) -> Possible[T]:
+    def _check_sequence(self) -> PossibleResult[T]:
         """Checks whether a result is a Sequence type.
 
         Catches generic sequences. All sequence types that are treated
@@ -228,7 +266,9 @@ class Deserialize(Generic[T]):  # pylint: disable=too-many-instance-attributes
             self.constructor_origin, Sequence
         ):
             if not isinstance(self.obj, Sequence):
-                raise DeserializeError(Sequence, self.obj, self.new_depth)
+                raise DeserializeError(
+                    Sequence, self.obj, self.new_depth, self.key
+                )
             if self.constructor_args:
                 _arg = self.constructor_args[0]
             else:
@@ -242,9 +282,9 @@ class Deserialize(Generic[T]):  # pylint: disable=too-many-instance-attributes
                 ).run()
                 for value in self.obj
             )  # type: ignore
-        return NoResult
+        return NO_RESULT
 
-    def _check_mapping(self) -> Possible[T]:
+    def _check_mapping(self) -> PossibleResult[T]:
         """Checks whether a result is a Mapping type.
 
         Catches generic mappings. All mapping types that are treated
@@ -254,7 +294,9 @@ class Deserialize(Generic[T]):  # pylint: disable=too-many-instance-attributes
             self.constructor_origin, Mapping
         ):
             if not isinstance(self.obj, Mapping):
-                raise DeserializeError(Mapping, self.obj, self.new_depth)
+                raise DeserializeError(
+                    Mapping, self.obj, self.new_depth, self.key
+                )
             if self.constructor_args:
                 _tpkey = self.constructor_args[0]
                 _tpvalue = self.constructor_args[1]
@@ -274,32 +316,36 @@ class Deserialize(Generic[T]):  # pylint: disable=too-many-instance-attributes
                         constructor=_tpvalue,
                         depth=self.new_depth,
                         convert_primitives=self.convert_primitives,
+                        key=key,
                     )
                     .run()
                     for key, value in self.obj.items()
                 }
             )  # type: ignore
-        return NoResult
+        return NO_RESULT
 
-    def _check_typed_dict(self) -> Possible[T]:
+    def _check_typed_dict(self) -> PossibleResult[T]:
         """Checks whether a result is a typing.TypedDict."""
         # pylint: disable=unidiomatic-typecheck
         if type(self.constructor) == _TypedDictMeta:
             # pylint: enable=unidiomatic-typecheck
             if not isinstance(self.obj, dict):
-                raise DeserializeError(dict, self.obj, self.new_depth)
+                raise DeserializeError(
+                    dict, self.obj, self.new_depth, self.key
+                )
             return {
                 name: Deserialize(
-                    obj=self.obj.get(name),
+                    obj=self.obj.get(name, UNDEFINED),
                     constructor=_type,
                     depth=self.new_depth,
                     convert_primitives=self.convert_primitives,
+                    key=name,
                 ).run()
                 for name, _type in get_type_hints(self.constructor).items()
             }  # type: ignore
-        return NoResult
+        return NO_RESULT
 
-    def _check_initvar_instance(self) -> Possible[T]:
+    def _check_initvar_instance(self) -> PossibleResult[T]:
         """Checks if a result is a dataclasses.InitVar."""
         if _is_initvar_instance(self.constructor):
             return Deserialize(
@@ -308,44 +354,70 @@ class Deserialize(Generic[T]):  # pylint: disable=too-many-instance-attributes
                 depth=self.new_depth,
                 convert_primitives=self.convert_primitives,
             ).run()
-        return NoResult
+        return NO_RESULT
 
-    def _check_none(self) -> Possible[T]:
+    def _check_none(self) -> PossibleResult[T]:
         """Checks if a result is None."""
         if self.constructor == type(None):
             if not self.obj is None:
-                raise DeserializeError(type(None), self.obj, self.new_depth)
+                raise DeserializeError(
+                    type(None), self.obj, self.new_depth, self.key
+                )
             return self.obj  # type: ignore
-        return NoResult
+        return NO_RESULT
 
-    def _check_primitive(self) -> Possible[T]:
+    def _check_undefined(self) -> PossibleResult[T]:
+        """Checks if a result is UNDEFINED.
+
+        This case is extremely rare / somewhat nonsensical, but is
+        included here for completeness sake.
+        """
+        if self.constructor == Undefined:
+            if not self.obj is UNDEFINED:
+                raise DeserializeError(
+                    Undefined, self.obj, self.new_depth, self.key
+                )
+            return self.obj  # type: ignore
+        return NO_RESULT
+
+    def _check_primitive(self) -> PossibleResult[T]:
         """Check if result is a primitive.
 
         If `self.convert_primitives` is `True`, tries to automatically
         convert primitive value to the specified type.
+
+        `None` and `UNDEFINED` are both considered "unconvertable"
         """
         if self.constructor in _PRIMITIVES:
+            if self.obj is UNDEFINED:
+                raise DeserializeError(
+                    self.constructor, self.obj, self.new_depth, self.key
+                )
+            if self.obj is None:
+                raise DeserializeError(
+                    self.constructor, self.obj, self.new_depth, self.key
+                )
             if not isinstance(self.obj, self.constructor):
                 if not self.convert_primitives:
                     raise DeserializeError(
-                        self.constructor, self.obj, self.new_depth
+                        self.constructor, self.obj, self.new_depth, self.key
                     )
                 try:
                     return self.constructor(self.obj)  # type: ignore
                 except (ValueError, TypeError) as error:
                     raise DeserializeError(
-                        self.constructor, self.obj, self.new_depth
+                        self.constructor, self.obj, self.new_depth, self.key
                     ) from error
             return self.obj
-        return NoResult
+        return NO_RESULT
 
-    def _check_any(self) -> Possible[T]:
+    def _check_any(self) -> PossibleResult[T]:
         """Check if result is typing.Any."""
         if _is_any(self.constructor):
             return self.obj  # type: ignore
-        return NoResult
+        return NO_RESULT
 
-    def _check_union(self) -> Possible[T]:
+    def _check_union(self) -> PossibleResult[T]:
         """Check if result is a Union.
 
         Encountering this type usually recursively disables
@@ -356,13 +428,15 @@ class Deserialize(Generic[T]):  # pylint: disable=too-many-instance-attributes
         if _is_union(self.constructor):
             args = get_args(self.constructor)
             is_optional = len(args) == 2 and type(None) in args
+            is_optional_property = len(args) == 2 and Undefined in args
             if is_optional and self.obj is None:
                 return None  # type: ignore
+            if is_optional_property and self.obj is UNDEFINED:
+                return UNDEFINED  # type: ignore
             for argument in args:
-                convert_primitives = (
-                    self.convert_primitives
-                    and is_optional
-                    and argument != type(None)
+                convert_primitives = self.convert_primitives and (
+                    (is_optional and argument != type(None))
+                    or (is_optional_property and argument != Undefined)
                 )
                 try:
                     return Deserialize(
@@ -373,10 +447,12 @@ class Deserialize(Generic[T]):  # pylint: disable=too-many-instance-attributes
                     ).run()
                 except DeserializeError:
                     pass
-            raise DeserializeError(self.constructor, self.obj, self.new_depth)
-        return NoResult
+            raise DeserializeError(
+                self.constructor, self.obj, self.new_depth, self.key
+            )
+        return NO_RESULT
 
-    def _check_typevar(self) -> Possible[T]:
+    def _check_typevar(self) -> PossibleResult[T]:
         """Checks whether it's a typevar."""
         if _is_typevar(self.constructor):  # type: ignore
             return Deserialize(
@@ -389,7 +465,7 @@ class Deserialize(Generic[T]):  # pylint: disable=too-many-instance-attributes
                 depth=self.new_depth,
                 convert_primitives=self.convert_primitives,
             ).run()
-        return NoResult
+        return NO_RESULT
 
 
 def _is_initvar_instance(typeval: Type) -> bool:
@@ -422,3 +498,5 @@ _TYPE_UNSAFE_CHECKS = (
 _PRIMITIVES = {str, int, float, bool}
 
 _ANY = {Any, object, InitVar}
+
+_UNION_PRIMITIVE_ALLOW_CONVERSION = {type(None), Undefined}
